﻿using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Problems.MultiObjectiveTestFunctions.Comparators;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {


  /// <summary>
  /// Spacing is defined as the standarddeviation of all d[i] 
  /// where d[i] is the minimum eukildean distance any point has 
  /// to all OTHER points in the same front 
  /// </summary>
  public class Spacing : IMultiObjectiveDistance {

    public Spacing() { }


    public static double GetSpacing(IEnumerable<double[]> front) {
      return new Spacing().Get(front);
    }
    public static double GetDistance(IEnumerable<double[]> front, IEnumerable<double[]> optimalFront) {
      return new Spacing().Get(front);
    }

    public double Compare(IEnumerable<double[]> front, IEnumerable<double[]> optimalFront) {
      return GetSpacing(front) - GetSpacing(optimalFront);
    }

    public double Get(IEnumerable<double[]> front) {
      //TODO build a kd-tree, sort the array, do someting intelligent here
      if (front == null) throw new ArgumentException("Fronts must not be null");
      List<double> d = new List<double>();
      foreach (double[] r in front) {
        double dist = Utilities.minDistance(r, front, false);
        d.Add(dist>=0?dist:0);
      }
      int n = d.Count;
      if (n == 0) throw new ArgumentException("Fronts must not be empty");
      return Math.Sqrt(d.Variance()*(n-1)/n);

    }

  }
}
