﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Encodings.RealVectorEncoding;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions.Testfunctions {

  /// <summary>
  /// unused class
  /// </summary>
  class Evaluators {

    //http://www.tik.ee.ethz.ch/sop/download/supplementary/testproblems/ [30.11.2015]
    private double[] ZDT1(RealVector r) {  //0,1
      double g = 0;
      for (int i = 1; i < r.Length; i++) g += r[i];
      g = 1.0 + 9.0 * g / (r.Length - 1);
      double[] res = { r[0], g * (1.0 - Math.Sqrt(r[0] / g)) };
      return res;
    }
    private double[] ZDT2(RealVector r) { //0,1
      double g = 0;
      for (int i = 1; i < r.Length; i++) g += r[i];
      g = 1.0 + 9.0 * g / (r.Length - 1);
      double d = r[0] / g;
      double[] res = { r[0], g * (1.0 - d * d) };
      return res;
    }
    private double[] ZDT3(RealVector r) { //0,1
      double g = 0;
      for (int i = 1; i < r.Length; i++) g += r[i];
      g = 1.0 + 9.0 * g / (r.Length - 1);
      double d = r[0] / g;
      double[] res = { r[0], g * (1.0 - Math.Sqrt(d) - d * Math.Sin(10 * Math.PI * r[0])) };
      return res;
    }
    private double[] ZDT4(RealVector r) {  //-5,5
      double g = 0;
      for (int i = 1; i < r.Length; i++) {
        double v = r[i];
        g += v * v - 10 * Math.Cos(4 * Math.PI * v);
      }
      g = 1.0 + 10.0 * (r.Length - 1) + g;
      double d = r[0] / g;
      double[] res = { r[0], g * (1.0 - d * d) };
      return res;
    }
    private double[] ZDT6(RealVector r) { //0,1
      double g = 0;
      for (int i = 1; i < r.Length; i++) g += r[i];
      g = g = 1.0 + 9.0 * Math.Pow(g / (r.Length - 1), 0.25);
      double d = r[0] / g;
      double[] res = { 1 - Math.Exp(-4 * r[0]) * Math.Pow(Math.Sin(6 * Math.PI * r[0]), 6), g * (1.0 - d * d) };
      return res;
    }


    //IMPOTANT There is often confusion about the numbering and the function definitions
    //eg.: in EALib only the first 4 functions are the same although the other 3 share resemblances
    //even on  http://www.tik.ee.ethz.ch/ where the original paper is referenced DTLZ7 on the page is DTLZ6 in the paper 
    /*Deb, K., Thiele, L., Laumanns, M., & Zitzler, E. (2002, May).
    Scalable multi-objective optimization test problems. 
    In Proceedings of the Congress on Evolutionary Computation (CEC-2002),(Honolulu, USA) (pp. 825-830). 
    http://repository.ias.ac.in/81671/ [30.11.15]
      */
    private double[] DTLZ1(RealVector r, int objectives) { //0,1
      double[] res = new double[objectives];

      //calculate g(Xm)
      double sum = 0, length = 0;
      for (int i = objectives; i < r.Length; i++) {
        double d = r[i] - 0.5;
        sum += d * d - Math.Cos(20 * Math.PI * d);
        length += r[i] * r[i];
      }
      length = Math.Sqrt(length);
      double g = 100 * (length + sum);

      //calculating f0...fM-1
      for (int i = 0; i < objectives; i++) {
        double f = 0.5 * i == 0 ? 1 : (1 - r[objectives - i - 1]) * (1 + g);
        for (int j = 0; j < objectives - i - 1; j++) {
          f *= r[j];
        }
        res[i] = f;
      }
      return res;
    }
    private double[] DTLZ2(RealVector r, int objectives) { //0,1
      double[] res = new double[objectives];

      //calculate g(Xm)
      double g = 0;
      for (int i = objectives; i < r.Length; i++) {
        double d = r[i] - 0.5;
        g += d * d;
      }

      //calculating f0...fM-1
      for (int i = 0; i < objectives; i++) {
        double f = i == 0 ? 1 : (Math.Sin(r[objectives - i - 1] * Math.PI / 2)) * (1 + g);
        for (int j = 0; j < objectives - i - 1; j++) {
          f *= Math.Cos(r[j] * Math.PI / 2);
        }
        res[i] = f;
      }
      return res;
    }
    private double[] DTLZ3(RealVector r, int objectives) {  //0,1
      double[] res = new double[objectives];

      //calculate g(Xm)
      double sum = 0, length = 0;
      for (int i = objectives; i < r.Length; i++) {
        double d = r[i] - 0.5;
        sum += d * d - Math.Cos(20 * Math.PI * d);
        length += r[i] * r[i];
      }
      length = Math.Sqrt(length);
      double g = 100 * (length + sum);

      //calculating f0...fM-1
      for (int i = 0; i < objectives; i++) {
        double f = i == 0 ? 1 : (Math.Sin(r[objectives - i - 1] * Math.PI / 2)) * (1 + g);
        for (int j = 0; j < objectives - i - 1; j++) {
          f *= Math.Cos(r[j] * Math.PI / 2);
        }
        res[i] = f;
      }
      return res;
    }
    private double[] DTLZ4(RealVector r, int objectives) { //0,1
      double[] res = new double[objectives];

      //calculate g(Xm)
      double g = 0;
      for (int i = objectives; i < r.Length; i++) {
        double d = r[i] - 0.5;
        g += d * d;
      }

      //calculating f0...fM-1
      for (int i = 0; i < objectives; i++) {
        double f = i == 0 ? 1 : (Math.Sin(Math.Pow(r[objectives - i - 1], 100) * Math.PI / 2)) * (1 + g);
        for (int j = 0; j < objectives - i - 1; j++) {
          f *= Math.Cos(Math.Pow(r[j], 100) * Math.PI / 2);
        }
        res[i] = f;
      }
      return res;
    }
    private double[] DTLZ5(RealVector r, int objectives) {  //0,1
      double[] res = new double[objectives];

      //calculate g(Xm)
      double g = 0;
      for (int i = objectives; i < r.Length; i++) {
        g += Math.Pow(r[i], 0.1);
      }

      //phi definition
      Func<double, double> phi;
      phi = (double x) => { return Math.PI / (4 * 1 + g) * (1 + 2 * g * x); };

      //calculating f0...fM-1
      for (int i = 0; i < objectives; i++) {
        double f = i == 0 ? 1 : (Math.Sin(phi(r[objectives - i - 1]) * Math.PI / 2)) * (1 + g);
        for (int j = 0; j < objectives - i - 1; j++) {
          f *= Math.Cos(phi(r[j]) * Math.PI / 2);
        }
        res[i] = f;
      }
      return res;
    }
    private double[] DTLZ6(RealVector r, int objectives) { //0,1
      double[] res = new double[objectives];

      //calculate g(Xm)
      double g = 0, length = 0;
      for (int i = objectives; i < r.Length; i++) {
        g += r[i];
        length += r[i] * r[i];
      }
      length = Math.Sqrt(length);
      g = 1.0 + 9.0 / length * g;

      //calculating f0...fM-2
      for (int i = 0; i < objectives - 1; i++) {
        res[i] = r[i];
      }
      //calculate fM-1
      double h = objectives;
      for (int i = 0; i < objectives - 1; i++) {
        h -= res[i] / (1 + g) * (1 + Math.Sin(3 * Math.PI * res[i]));
      }
      res[objectives - 1] = (1 + g) * h;

      return res;
    }
    private double[] DTLZ7(RealVector r, int objectives) { //0,1
      double[] res = new double[objectives];

      //calculate f0...fM-1;
      double n = 10 * objectives;
      for (int i = 0; i < objectives; i++) {
        double d = 0;
        for (int j = (int)Math.Floor((i - 1) * n / objectives); j < (int)Math.Floor(i * n / objectives); j++) {
          d += r[j];
        }
        d *= 1 / Math.Floor(n / objectives);
        res[i] = d;
      }

      //evaluate constraints g0...gM-2
      for (int i = 0; i < objectives - 1; i++) {
        if (res[objectives - 1] + 4 * res[i] - 1 < 0) return null; //TODO null is not the way to go
      }
      //evaluate gM-1
      double min = Double.PositiveInfinity;
      for (int i = 0; i < objectives - 1; i++) {
        for (int j = 0; j < i; j++) min = Math.Min(min, res[i] + res[j]);
      };
      if (2 * res[objectives - 1] + min - 1 < 0) return null;  //TODO null is not the way to go

      return res;
    }

    //original paper not accessible + wikipedia definition works only with fixed (3) dimensions
    //http://darwin.di.uminho.pt/jecoli/index.php/Multiobjective_example [30.11.2015]
    private double[] Kursawe(RealVector r) { //-5,5
      //objective 1
      double f0 = 0.0;
      for (int i = 0; i < r.Length - 1; i++) {
        f0 += -10 * Math.Exp(-0.2 * Math.Sqrt(r[i] * r[i] + r[i + 1] * r[i + 1]));
      }
      //objective2
      double f1 = 0.0;
      for (int i = 0; i < r.Length; i++) {
        f1 += Math.Pow(Math.Abs(r[i]), 0.2) + 5 * Math.Sin(Math.Pow(r[i], 3));
      }

      double[] res = { f0, f1 };
      return res;
    }
    //http://www.eng.buffalo.edu/Research/MODEL/mdo.test.orig/class2prob4/descr.html
    //TODO this is not multi-objective
    private double[] Golinski(RealVector r) { //bounds {{2.6,3.6}, {0.7,0.8}, {17,28}, {7.3, 8.3}, {7.3,8.3}, {2.9,3.9}, {5.0,5.5}}
      if (r.Length != 7) return null;

      double[] Cf = { 0.7854, 3.3333, 14.9334, 43.0934, 1.5079, 7.477 };
      double f = Cf[0] * r[0] * r[1] * r[1];
      f*= Cf[1] * r[2] * r[2] + Cf[2] * r[2] - Cf[3];
      f -= Cf[4] * (r[5] * r[5] + r[6] * r[6]);
      f += Cf[5] * (r[5] * r[5] * r[5] + r[6] * r[6] * r[6]);
      f += Cf[0] * (r[3] * r[5] * r[5] + r[4] * r[6] * r[6]);

      double Ca12s = 745.0*745.0; 
      //Constraints x[0],x[1],x[2],x[5],x[6]
      List<GoldinskiConstraint> constraints = new List<GoldinskiConstraint>();
      constraints.Add(new GoldinskiConstraint(x => 27.0/(x[0]*x[1]*x[1]*x[2])));
      constraints.Add(new GoldinskiConstraint(x => 397.5 / (x[0] * x[1] * x[1] * x[2]*x[2])));
      constraints.Add(new GoldinskiConstraint(x => 1.93*Math.Pow(x[3],3) / (x[1] * x[2] * Math.Pow(x[5],4))));
      constraints.Add(new GoldinskiConstraint(x => 1.93 * Math.Pow(x[4], 3) / (x[1] * x[2] * Math.Pow(x[6], 4))));
      constraints.Add(new GoldinskiConstraint(x => Math.Sqrt(Ca12s*x[3]*x[3]/(x[1]*x[1]*x[2]*x[2])+0.169e8)/(1100*0.1*Math.Pow(x[5],3))));
      constraints.Add(new GoldinskiConstraint(x => Math.Sqrt(Ca12s * x[4] * x[4] / (x[1] * x[1] * x[2] * x[2]) + 0.169e8) / (859 * 0.1 * Math.Pow(x[6], 3))));
      constraints.Add(new GoldinskiConstraint(x => x[1]*x[2]/40.0));
      constraints.Add(new GoldinskiConstraint(x => 5.0 * x[1] / x[0]));
      constraints.Add(new GoldinskiConstraint(x => x[0] / 12 * x[1]));
      constraints.Add(new GoldinskiConstraint(x => 2.6/x[0])); //g10
      constraints.Add(new GoldinskiConstraint(x => x[0]/3.6));
      constraints.Add(new GoldinskiConstraint(x => 0.7 / x[1]));
      constraints.Add(new GoldinskiConstraint(x => x[1] / 0.8));
      return null;
    }
    private class GoldinskiConstraint {
      private Func<RealVector, double> phi;
      public GoldinskiConstraint(Func<RealVector, double> phi) {
        this.phi = phi;
      }
      public bool eval(RealVector r) {
        return phi(r) <= 1.0;
      }
    } 


    // https://en.wikipedia.org/wiki/Test_functions_for_optimization [30.11.2015]
    private double[] Fonseca(RealVector r) {  //-4,4
      double f0 = 0.0, aux = 1.0 / Math.Sqrt(r.Length);

      //objective1
      for (int i = 0; i < r.Length; i++) {
        double d = r[i] - aux;
        f0 += d * d;
      }
      f0 = 1 - Math.Exp(-f0);

      //objective2
      double f1 = 0.0;
      for (int i = 0; i < r.Length; i++) {
        double d = r[i] + aux;
        f1 += d * d;
      }
      f1 = 1 - Math.Exp(-f1);

      double[] res = { f0, f1 };
      return res;
    }
    private double[] Schaffer1(RealVector r) { //-A,A
      if (r.Length != 1) return null;
      double x = r[0];

      //objective1
      double f0=x*x;

      //objective0
      double f1 = x - 2;
      f1 *= f1;

      double[] res = { f0, f1 };
      return res;
    }
    private double[] Schaffer2(RealVector r) { //-5,10
      if (r.Length != 1) return null;
      double x = r[0];

      //objective1
      double f0;
      if (x <= 11) f0 = -x;
      else if (x <= 3) f0 = x - 2;
      else if (x <= 4) f0 = 4 - x;
      else f0 = x - 4;

      //objective0
      double f1 = x - 5;
      f1 *= f1;

      double[] res = { f0, f1 };
      return res;
    }

  }
}
