﻿using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MultiObjectiveTestFunctions {
  [Item("Kursawe", "Kursawe function from // http://darwin.di.uminho.pt/jecoli/index.php/Multiobjective_example [30.11.2015]")]
  [StorableClass]
  public class Kursawe : MultiObjectiveTestFunction {

    public override double[,] Bounds(int objectives) {
      return new double[,] { { -5, 5 } };
    }

    public override bool[] Maximization(int objecitves) {
      return new bool[2];
    }

    public override int MinimumObjectives {
      get { return 2; }
    }
    public override int MaximumObjectives {
      get {return 2;}
    }

    public override int MinimumSolutionLength {
      get {return 3;}
    }
    public override int MaximumSolutionLength {
      get { return int.MaxValue; }
    }

    public override IEnumerable<double[]> OptimalParetoFront(int objecitves) {
        return PFStore.get(this.ItemName);
    }
    public override double BestKnownHypervolume(int objecitves) {
        return new Hypervolume(ReferencePoint(objecitves), Maximization(2)).GetHypervolume(OptimalParetoFront(objecitves));
    }
    public override double[] ReferencePoint(int objectives) {
      return new double[] { 11, 11 };
    }

    [StorableConstructor]
    protected Kursawe(bool deserializing) : base(deserializing) { }
    protected Kursawe(Kursawe original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new Kursawe(this, cloner);
    }
    public Kursawe() : base() { }

    



    public override double[] Evaluate(RealVector r, int objectives) {
      if (objectives != 2) throw new ArgumentException("The Kursawe problem must always have 2 objectives");
      //objective 1
      double f0 = 0.0;
      for (int i = 0; i < r.Length-1; i++) {
        f0 += -10 * Math.Exp(-0.2 * Math.Sqrt(r[i] * r[i] + r[i + 1] * r[i + 1]));
      }
      //objective2
      double f1 = 0.0;
      for (int i = 0; i < r.Length; i++) {
        f1 += Math.Pow(Math.Abs(r[i]), 0.8) + 5 * Math.Sin(Math.Pow(r[i], 3));
      }

      return new double[] { f0, f1 };
    }
  }
}
